#!/usr/bin/env python

import optparse
import os
import shutil
import sys
import unittest

USER = os.environ.get('USER') or 'root'


def runtests(suite, verbosity=1, failfast=False):
    runner = unittest.TextTestRunner(verbosity=verbosity, failfast=failfast)
    results = runner.run(suite)
    return results.failures, results.errors

def get_option_parser():
    usage = 'usage: %prog module1, module2 ...'
    parser = optparse.OptionParser(usage=usage)
    basic = optparse.OptionGroup(parser, 'Basic test options')
    basic.add_option('-v', '--verbosity', dest='verbosity', default=1,
                     type='int', help='Verbosity of output')
    basic.add_option('-f', '--failfast', action='store_true', default=False,
                     dest='failfast', help='Exit on first failure/error.')
    parser.add_option_group(basic)
    return parser

def collect_tests(args):
    suite = unittest.TestSuite()

    if not args:
        import tests
        module_suite = unittest.TestLoader().loadTestsFromModule(tests)
        suite.addTest(module_suite)
    else:
        cleaned = ['tests.%s' % arg if not arg.startswith('tests.') else arg
                   for arg in args]
        user_suite = unittest.TestLoader().loadTestsFromNames(cleaned)
        suite.addTest(user_suite)

    return suite

if __name__ == '__main__':
    parser = get_option_parser()
    options, args = parser.parse_args()

    suite = collect_tests(args)
    failures, errors = runtests(suite, options.verbosity, options.failfast)

    if errors:
        sys.exit(2)
    elif failures:
        sys.exit(1)

    sys.exit(0)



#-----------------------------------------------------------------------------
# Monkey patching of clingo for the clorm library
# ------------------------------------------------------------------------------

import sys
import clingo as original_clingo
from . import clingo as clorm_clingo

original_control = original_clingo.Control

#------------------------------------------------------------------------------
# clingo patching and unpatching
#------------------------------------------------------------------------------

def patch():
    """Patch clingo to add interface for Clorm objects.

    This patch replaces clingo.Control with
    clorm.clingo.Control. clorm.clingo.Control wraps the clingo.Control
    functions to add interfaces for Clorm fact and factbase objects. On solving
    it returns clorm.clingo.Model or clorm.clingo.SolveHandle object (as
    necessary).
    """
    original_clingo.Control=clorm_clingo.Control


def unpatch():
    """Reverse the patching of clingo."""
    original_clingo.Control=original_control

#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')




# -----------------------------------------------------------------------------
# Functions and classes to help with converting symbols to facts and
# vice-versa. This includes the unification process of unifying a set of symbols
# to matching predicate signatures as well as more high-level tasks such as
# asserting a set of facts into the solver (ie. clingo backend)
#
# Also includes SymbolPredicateUnifier and unify function
# implementation. Although should refactor SymbolPredicateUnifier to something
# less heavy handed.
# ------------------------------------------------------------------------------

import itertools
import clingo
import clingo.ast as clast
from .core import *
from .noclingo import is_Function, is_Number, SymbolMode
from .factbase import *
from .core import get_field_definition, PredicatePath, kwargs_check_keys, \
    get_symbol_mode


__all__ = [
    'SymbolPredicateUnifier',
    'unify',
    'control_add_facts',
    'symbolic_atoms_to_facts',
    'parse_fact_string',
    'parse_fact_files',
    'UnifierNoMatchError',
    'NonFactError'
    ]

#------------------------------------------------------------------------------
# Global
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Clorm exception subclasses
#------------------------------------------------------------------------------

class UnifierNoMatchError(ClormError):
    def __init__(self,message,symbol,predicates):
        super().__init__(message)
        self._symbol=symbol
        self._predicates=predicates
    @property
    def symbol(self): return self._symbol
    @property
    def predicates(self): return self._predicates

class NonFactError(ClormError):
    def __init__(self,message):
        super().__init__(message)

#------------------------------------------------------------------------------
# A unifier takes a list of predicates to unify against (order matters) and a
# set of raw clingo symbols against this list.
# ------------------------------------------------------------------------------

class Unifier(object):
    def __init__(self,predicates=[]):
        self._predicates=tuple(predicates)
        self._pgroups = {}
        self._add_predicates(predicates)

    def _add_predicates(self,predicates=[]):
        for p in predicates:
            self._pgroups.setdefault((p.meta.name,p.meta.arity),[]).append(p)

    def add_predicate(self,predicate):
        self._add_predicates([predicates])

    def unify_symbol(self,sym,*,raise_nomatch=False):
        mpredicates = self._pgroups.get((sym.name,len(sym.arguments)),[])
        for pred in mpredicates:
            try:
                return pred._unify(sym)
            except ValueError:
                pass
        if raise_nomatch:
            raise UnifierNoMatchError(
                f"Cannot unify symbol '{sym}' to predicates in {self._predicates}",
                sym, self._predicates)
        return None

    def unify(self,symbols,*,factbase=None,raise_nomatch=False):
        fb=FactBase() if factbase is None else factbase
        for sym in symbols:
            f=self.unify_symbol(sym,raise_nomatch=raise_nomatch)
            if f is not None: fb.add(f)

        return fb

#------------------------------------------------------------------------------
# A fact generator that takes a list of predicates to unify against (order
# matters) and a set of raw clingo symbols against this list.
# ------------------------------------------------------------------------------

def _unify(predicates, symbols):
    def unify_single(cls, r):
        try:
            return cls._unify(r)
        except ValueError:
            return None

    # To make things a little more efficient use the name/arity signature as a
    # filter. However, Python doesn't have a built in multidict class, and I
    # don't want to add a dependency to an external library just for one small
    # feature, so implement a simple structure here.
    sigs = [((cls.meta.name, len(cls.meta)),cls) for cls in predicates]
    types = {}
    for sig,cls in sigs:
        if sig not in types: types[sig] = [cls]
        else: types[sig].append(cls)

    # Loop through symbols and yield when we have a match
    for raw in symbols:
        classes = types.get((raw.name, len(raw.arguments)))
        if not classes: continue
        for cls in classes:
            f = unify_single(cls,raw)
            if f is not None:
                yield f
                break


#------------------------------------------------------------------------------
# SymbolPredicateUnifier offers a decorator interface for gathering predicate and index
# definitions to be used in defining a FactBase subclass.
# ------------------------------------------------------------------------------
class SymbolPredicateUnifier(object):
    """A fact base builder simplifies the task of unifying raw clingo.Symbol objects
    with Clorm predicates. Predicates classes are registered using the
    'register' function (which can be called as a normal function or as a class
    decorator.
    """

    def __init__(self, predicates=[], indexes=[], suppress_auto_index=False):
        self._predicates = ()
        self._indexes = ()
        self._suppress_auto_index = suppress_auto_index
        tmppreds = []
        tmpinds = []
        tmppredset = set()
        tmpindset = set()
        for pred in predicates:
                self._register_predicate(pred,tmppreds,tmpinds,tmppredset,tmpindset)
        for fld in indexes:
                self._register_index(fld,tmppreds,tmpinds,tmppredset,tmpindset)
        self._predicates = tuple(tmppreds)
        self._indexes = tuple(tmpinds)

    def _register_predicate(self, cls, predicates, indexes, predicateset, indexset):
        if not issubclass(cls, Predicate):
            raise TypeError("{} is not a Predicate sub-class".format(cls))
        if cls in predicateset: return
        predicates.append(cls)
        predicateset.add(cls)
        if self._suppress_auto_index: return

        # Add all fields (and sub-fields) that are specified as indexed
        for fp in cls.meta.indexes:
            self._register_index(fp,predicates,indexes,predicateset,indexset)

    def _register_index(self, path, predicates, indexes, predicateset, indexset):
        if path.meta.hashable in indexset: return
        if isinstance(path, PredicatePath) and path.meta.predicate in predicateset:
            indexset.add(path.meta.hashable)
            indexes.append(path)
        else:
            raise TypeError("{} is not a predicate field for one of {}".format(
                path, [ p.__name__ for p in predicates ]))

    def register(self, cls):
        if cls in self._predicates: return cls
        predicates = list(self._predicates)
        indexes = list(self._indexes)
        tmppredset = set(self._predicates)
        tmpindset = set([p.meta.hashable for p in self._indexes])
        self._register_predicate(cls,predicates,indexes,tmppredset,tmpindset)
        self._predicates = tuple(predicates)
        self._indexes = tuple(indexes)
        return cls

    def unify(self, symbols, delayed_init=False, raise_on_empty=False):
        def _populate():
            facts=list(_unify(self.predicates, symbols))
            if not facts and raise_on_empty:
                raise ValueError("FactBase creation: failed to unify any symbols")
            return facts

        if delayed_init:
            return FactBase(facts=_populate, indexes=self._indexes)
        else:
            return FactBase(facts=_populate(), indexes=self._indexes)

    @property
    def predicates(self): return self._predicates
    @property
    def indexes(self): return self._indexes

#------------------------------------------------------------------------------
# Generate facts from an input array of Symbols.  The `unifier` argument takes a
# list of predicate classes or a SymbolPredicateUnifer object to unify against
# the symbol object contained in `symbols`.
# ------------------------------------------------------------------------------

def unify(unifier,symbols,ordered=False):
    '''Unify raw symbols against a list of predicates or a SymbolPredicateUnifier.

    Symbols are tested against each predicate unifier until a match is
    found. Since it is possible to define multiple predicate types that can
    unify with the same symbol, the order of the predicates in the unifier
    matters. With the `ordered` option set to `True` a list is returned that
    preserves the order of the input symbols.

    Args:
      unifier: a list of predicate classes or a SymbolPredicateUnifier object.
      symbols: the symbols to unify.
      ordered (default: False): optional to return a list rather than a FactBase.
    Return:
      a FactBase containing the unified facts, indexed by any specified indexes,
         or a list if the ordered option is specified

    '''
    if not unifier:
        raise ValueError(("The unifier must be a list of predicates "
                          "or a SymbolPredicateUnifier"))
    if ordered:
        if isinstance(unifier, SymbolPredicateUnifier):
            unifier=unifier.predicates
        return list(_unify(unifier,symbols))
    else:
        if not isinstance(unifier, SymbolPredicateUnifier):
            unifier=SymbolPredicateUnifier(predicates=unifier)
        return unifier.unify(symbols)

#------------------------------------------------------------------------------
# Function to add a collection of facts to the solver backend
#------------------------------------------------------------------------------

if clingo.__version__ >= "5.5.0":

    def control_add_facts(ctrl, facts):
        with ctrl.backend() as bknd:
            for f in facts:
                raw=f.raw if isinstance(f,Predicate) else f
                atm = bknd.add_atom(raw)
                bknd.add_rule([atm])
else:
    import clingo.ast as ast

    def control_add_facts(ctrl, facts):
        with ctrl.builder() as bldr:
            line=1
            for f in facts:
                raw=f.raw if isinstance(f,Predicate) else f
                floc = { "filename" : "<input>", "line" : line , "column" : 1 }
                location = { "begin" : floc, "end" : floc }
                r = ast.Rule(location,
                             ast.Literal(location, ast.Sign.NoSign,
                                         ast.SymbolicAtom(ast.Symbol(location,raw))),
                             [])
                bldr.add(r)
                line += 1

control_add_facts.__doc__ = '''Assert a collection of facts to the solver

    Provides a flexible approach to asserting facts to the solver. The facts can
    be either `clingo.Symbol` objects or clorm facts and can be in any type of
    be in any collection (including a `clorm.FactBase`).

    Args:
        ctrl: a `clingo.Control` object
        facts: the collection of facts to be asserted into the solver
'''

# ------------------------------------------------------------------------------
# Function to take a SymbolicAtoms object and extract facts from it
# ------------------------------------------------------------------------------

def symbolic_atoms_to_facts(symbolic_atoms, unifier, *,
                            facts_only=False, factbase=None):
    '''Extract `clorm.FactBase` from `clingo.SymbolicAtoms`

    A `clingo.SymbolicAtoms` object is returned from the
    `clingo.Control.symbolic_atoms` property. This property is a view into the
    internal atoms within the solver and can be examined at anytime
    (ie. before/after the grounding/solving). Some of the atoms are trivially
    true (as determined by the grounder) while others may only be true in some
    models.

    Args:
        symbolic_atoms: a `clingo.SymbolicAtoms` object
        unifier: a list of Clorm Predicate sub-classes to unify against
        facts_only (default False): return facts only or include contingent literals
        factbase (default None): add to existing FactBase or return a new FactBase
    '''

    if factbase is None: factbase=FactBase()

    def group_predicates():
        groups = {}
        for pcls in unifier:
            groups.setdefault((pcls.meta.name,pcls.meta.arity),[]).append(pcls)
        return groups

    def single(cls,sym):
        try:
            factbase.add(cls._unify(sym))
            return True
        except ValueError:
            return False

    groups = group_predicates()
    for (name,arity), mpredicates in groups.items():
        for symatom in itertools.chain(
                symbolic_atoms.by_signature(name,arity,True),
                symbolic_atoms.by_signature(name,arity,False)):
            if facts_only and not symatom.is_fact: continue
            for pcls in mpredicates:
                if single(pcls,symatom.symbol): continue
    return factbase


#------------------------------------------------------------------------------
# parse a string containing ASP facts into a factbase
#------------------------------------------------------------------------------

def _t2d(t):
    return { k : v for k,v in t }

def _negate(s):
    if is_Number(s):
        return symbols.Number(s.number*-1)
    if is_Function(s):
        return symbols.Function(s.name,s.arguments,not s.positive)
    raise NonFactError(f"'{s}' cannot be negated in a simple fact")

class _ASTVisitor(object):
    def __init__(self,unifier,*,factbase=None,
                 raise_nomatch=False,raise_nonfact=False):
        self._stack = []
        self._location=None
        self._unifier = unifier
        self._fb = FactBase() if factbase is None else factbase
        self._raise_nomatch=raise_nomatch
        self._raise_nonfact=raise_nonfact
        self._stmtstr = ""

    def _err(self,msg):
        if self._raise_nonfact:
            raise NonFactError(f"{msg} in '{self._stmtstr}' {self._location}")
        return None

    def symbolic_term(self,x):
        items=_t2d(x.items())
        self._location=items['location']
        symbol=items['symbol']
        self._stack.append(symbol)

    def unary_operation(self,x):
        items=_t2d(x.items())
        self._location=items['location']
        otype=items['operator_type']
        if otype != clast.UnaryOperator.Minus:
            return self._err(f"Unexpected non minus unary operator in simple fact")
        argument=items['argument']
        self.term(argument)
        self._stack.append(_negate(self._stack.pop()))

    def function(self,x):
        items=_t2d(x.items())
        self._location=items['location']
        name=items['name']
        arguments=items['arguments']
        external=items['external']

        if external:
            return self._err(f"'{x}' is an external function")
        if arguments:
            for a in arguments: self.term(a)
            args=self._stack[-1*len(arguments):]
            del self._stack[-1*len(arguments):]
        else:
            args=[]
        tmp=symbols.Function(name,args)
        self._stack.append(tmp)

    def term(self,x):
        if x.ast_type == clast.ASTType.SymbolicTerm:
            self.symbolic_term(x)
        elif x.ast_type == clast.ASTType.UnaryOperation:
            self.unary_operation(x)
        elif x.ast_type == clast.ASTType.Function:
            self.function(x)
        else:
            return self._err(f"'{x}' ({x.ast_type}) is not a simple fact")

    def head(self,x):
        if x.ast_type == clast.ASTType.Disjunction:
            return self._err(f"Disjunction '{x}' is not a simple fact")
        if x.ast_type == clast.ASTType.Aggregate:
            return self._err(f"Aggregate '{x}' is not a simple fact")
        if x.ast_type != clast.ASTType.Literal:
            return self._err(f"'{x}' is not a simple fact literal")
        hitems=_t2d(x.items())
        self._location=hitems['location']
        sign=hitems['sign']
        atom=hitems['atom']

        aitems=_t2d(atom.items())
        if 'symbol' not in aitems:
            return self._err(f"'{atom}' is not a symbol")
        self.term(aitems['symbol'])

    def rule(self,x):
        items=_t2d(x.items())
        self._location=items['location']

        if items['body']: return self._err(f"Rule '{x}' is not a fact")
        self.head(items['head'])

    def stmt(self,ast):
        self._stmtstr = str(ast)
        if ast.ast_type != clast.ASTType.Rule: return None
        self.rule(ast)

    def facts(self):
        return self._unifier.unify(self._stack,factbase=self._fb,
                                   raise_nomatch=self._raise_nomatch)

#------------------------------------------------------------------------------
# Parse ASP facts from a string or files into a factbase
#------------------------------------------------------------------------------

def parse_fact_string(aspstr,unifier,*,factbase=None,
                      raise_nomatch=False,raise_nonfact=False):
    '''Parse a string of ASP facts into a FactBase

    Facts must be of a simple form that can correspond to clorm predicate
    instances. Rules that are NOT simple facts include: any rule with a body, a
    disjunctive fact, a choice rule, a theory atom, a literal with an external
    @-function reference, a literal that requires some mathematical calculation
    (eg., "p(1+1).")

    NOTE: Currently, this function is not safe when running in NOCLINGO mode and
    will raise a NotImplementedError if called.

    Args:
      aspstr: an ASP string containing the facts
      factbase: if no factbase is specified then create a new one
      unifier: a list of clorm.Predicate classes to unify against
      raise_nomatch: raise UnifierNoMatchError on a fact that cannot unify
      raise_nonfact: raise NonFactError on any non simple fact (eg. complex rules)

    '''

    if get_symbol_mode() == SymbolMode.NOCLINGO:
        raise NotImplementedError("Function not supported in NOCLINGO mode")

    v=_ASTVisitor(unifier=Unifier(unifier),factbase=factbase,
                  raise_nomatch=raise_nomatch,raise_nonfact=raise_nonfact)

    clast.parse_string(aspstr,v.stmt)
    return v.facts()

def parse_fact_files(files,unifier,*,factbase=None,
                     raise_nomatch=False,raise_nonfact=False):
    '''Parse the facts from a list of files into a FactBase

    Facts must be of a simple form that can correspond to clorm predicate
    instances. Rules that are NOT simple facts include: any rule with a body, a
    disjunctive fact, a choice rule, a theory atom, a literal with an external
    @-function reference, a literal that requires some mathematical calculation
    (eg., "p(1+1).")

    NOTE: Currently, this function is not safe when running in NOCLINGO mode and
    will raise a NotImplementedError if called.

    Args:
      files: a list of ASP files containing the facts
      factbase: if no factbase is specified then create a new one
      unifier: a list of clorm.Predicate classes to unify against
      raise_nomatch: raise UnifierNoMatchError on a fact that cannot unify
      raise_nonfact: raise NonFactError on any non simple fact (eg. complex rules)
    '''

    if get_symbol_mode() == SymbolMode.NOCLINGO:
        raise NotImplementedError("Function not supported in NOCLINGO mode")

    v=_ASTVisitor(unifier=Unifier(unifier),factbase=factbase,
                  raise_nomatch=raise_nomatch,raise_nonfact=raise_nonfact)

    clast.parse_files(files,v.stmt)
    return v.facts()


#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')



'''Functions and classes to wrap existing classes. Provides a wrapper metaclass
and also a function that returns a wrapped class. The function is more flexible
as a metaclass has multiple inheritence limitations.

A wrapper metaclass for building wrapper objects. It is instantiated by
specifying a class to be to be wrapped as a parent class with WrapperMetaClass
as the metaclass. This creates a wrapped/proxy base class for that type of
object. Note: this subverts the subclass mechanism as it does not actually
create a subclass of the wrapped class. Instead it is simply used to create the
forwarding of the member functions and attributes, while the wrapped class is
replaced with object as the parent.

Note: if a constructor is provided for the wrapper then it should call
init_wrapper manually. It also sets up '_wrapped' and '_wrapped_cls' attributes
so these cannot be attributes of the wrapped class.

This metaclass is to be used for wrapping clingo.Control, clingo.SolveHandle,
and clingo.Model objects.

Note: some ideas and code have been copied from:
https://code.activestate.com/recipes/496741-object-proxying/

'''

import functools
import inspect

# ------------------------------------------------------------------------------
# Make proxy member functions and properties
# ------------------------------------------------------------------------------
def _make_wrapper_function(fn):
    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        func=getattr(self._wrapped,fn.__name__)
        return func(*args, **kwargs)
    return wrapper

def _make_wrapper_property(name, get_only=True):
    def getter(self):
        return getattr(self._wrapped,name)
    def setter(self,x):
        return self._wrapped.__setattr__(name,x)
    return property(getter,setter)

def _check_wrapper_object(wrapper,strict=False):
    ActualType = type(wrapper._wrapped)
    WrappedType = wrapper._wrapped_cls
    if issubclass(ActualType,WrappedType): return
    if strict:
        raise TypeError(("Invalid proxied object {} not of expected type "
                         "{}").format(wrapper._wrapped,WrappedType))

# Constructor for every Predicate sub-class
def init_wrapper(wrapper, *args, **kwargs):
    Wrapped = wrapper._wrapped_cls
    if "wrapped_" in kwargs:
        if len(args) != 0 and len(kwargs) != 1:
            raise ValueError(("Invalid initialisation: the 'wrapped_' argument "
                              "cannot be combined with other arguments"))
        wrapper._wrapped = kwargs["wrapped_"]
        _check_wrapper_object(wrapper,strict=False)
    else:
        wrapper._wrapped = Wrapped(*args,**kwargs)

# ------------------------------------------------------------------------------
# The wrapper metaclass
# ------------------------------------------------------------------------------
class WrapperMetaClass(type):

    def __new__(meta, name, bases, dct):
        if len(bases) != 1:
            raise TypeError("ProxyMetaClass requires exactly one parent class")
        Wrapped = bases[0]
        bases = (object,)

        ignore=["__init__", "__new__", "__dict__", "__weakref__", "__setattr__",
                "__getattr__"]

        if "_wrapped_cls" in dct:
            raise TypeError(("ProxyMetaClass cannot proxy a class with a "
                             "\"_wrapped_cls\" attribute: {}").format(PrClass))
        dct["_wrapped_cls"] = Wrapped

        # Mirror the attributes of the proxied class
        for key,value in Wrapped.__dict__.items():
            if key in ignore: continue
            if key in dct: continue

            if callable(value):
                dct[key]=_make_wrapper_function(value)
            else:
                dct[key]=_make_wrapper_property(key)

        # Create the init function if none is provided
        if "__init__" not in dct: dct["__init__"] = init_wrapper

        return super(WrapperMetaClass, meta).__new__(meta, name, bases, dct)

#------------------------------------------------------------------------------
# Alternative wrapper implementation that doesn't use a metaclass. The metaclass
# version is a problem when wrapping a class that already has a metaclass.  This
# version takes a class to be wrapped and an optional override class. In then
# creates an wrapper class that has all the properties and member functions of
# the override class as well as the wrapped class (with the override class
# overriding any function/property that is common to both).
# ------------------------------------------------------------------------------

def make_class_wrapper(inputclass, override=None):
    def getattrdoc(cls, key):
        if not cls: return None
        try:
            attr = getattr(cls,key)
            if isinstance(attr,property) or callable(attr):
                return attr.__doc__
            return None
        except:
            return None

    Wrapped = inputclass
    name = inputclass.__name__

    w_ignore=set(["__init__", "__new__",  "__del__","__weakref__", "__setattr__",
                  "__getattr__", "__module__", "__name__", "__dict__",
                  "__abstractmethods__", "__orig_bases__", "__parameters__", "_abc_impl"])
    o_ignore=set(["__module__", "__new__", "__dict__", "__weakref__", "__name__"])
    dct = {}
    if override:
        for key,value in override.__dict__.items():
            if key in o_ignore: continue
            if key == "__doc__" and not value: continue
            dct[key] = value

    if "_wrapped_cls" in dct:
        raise TypeError(("The overrides cannot contain a "
                         "\"_wrapped_cls\" attribute: {}").format(dct))
    dct["_wrapped_cls"] = Wrapped

    # Mirror the attributes of the proxied class
    for key,value in Wrapped.__dict__.items():
        if key in w_ignore: continue
        if key in dct: continue

        if key == "__doc__" and key not in dct:
            dct[key]=value
        elif callable(value):
            dct[key]=_make_wrapper_function(value)
        else:
            dct[key]=_make_wrapper_property(key)

    # Create a basic init function if none is provided
    if "__init__" not in dct: dct["__init__"] = init_wrapper

    WrapperClass = type(name,(object,),dct)

#    print("\n{}".format(name))
#    if override:
#        print ("OVERRIDE: {}".format(override.__dict__.keys()))
#    print ("WRAPPED: {}\n".format(inputclass.__dict__.keys()))
#    print("NAME: {} : {}".format(WrapperClass.__name__, WrapperClass.__dict__.keys()))

    # Now go through and add docstrings if necessary
    for key in dir(WrapperClass):
        attr = getattr(WrapperClass, key)
        if inspect.isclass(attr): continue
        if callable(attr) and attr.__doc__: continue
        doc1 = getattrdoc(override,key)
        doc2 = getattrdoc(inputclass,key)
        if doc1: attr.__doc__ = doc1
        elif doc2: attr.__doc__ = doc2

    return WrapperClass


#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')




#--------------------------------------------------------------------------------
# JSON Encoder/Decoder for clorm.Predicate and clingo.Symbol objects
#
# Note: the JSON encoding of the clingo.Symbol objects is not the same as
# running clingo with the ``--outf=2`` argument. The output here is intended to
# be easily machine processable whereas the clingo output would require parsing
# to regenerate the original symbol objects.
# --------------------------------------------------------------------------------
from collections.abc import Mapping
import clingo
import json
from .orm import *

__all__ = [
    'symbol_encoder',
    'symbol_decoder',
    'FactBaseCoder'
]

def _raise(obj):
    otype = type(obj)
    raise TypeError("Object of type '{}' is not JSON serializable".format(otype))

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------

def symbol_encoder(s):
    '''A JSON encoder for clingo.Symbol objects.

    Example usage:
    sym = clingo.Function("afact", [clingo.Number(1)])
    json_str = json.dumps(sym, default=encoder)

    Args:
      symbol(clingo.Symbol): a symbol object
    '''

    if not isinstance(s, clingo.Symbol): _raise(s)

    def stypestr(stype):
        st2str = {
            clingo.SymbolType.Infimum : "Infimum",
            clingo.SymbolType.Supremum : "Supremum",
            clingo.SymbolType.String : "String",
            clingo.SymbolType.Number : "Number",
            clingo.SymbolType.Function : "Function"}
        return st2str[stype]

    # A clingo.Symbol object
    js = {}
    js["clingo.SymbolType"] = stypestr(s.type)
    if s.type == clingo.SymbolType.Infimum: return js
    if s.type == clingo.SymbolType.Supremum: return js
    if s.type == clingo.SymbolType.Number:
        js["number"] = s.number
        return js
    if s.type == clingo.SymbolType.String:
        js["string"] = s.string
        return js
    if s.type == clingo.SymbolType.Function:
        js["name"] = s.name
        js["arguments"] = [symbol_encoder(a) for a in s.arguments]
        js["positive"] = s.positive
        return js

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------

def symbol_decoder(obj):
    '''A JSON Decoder for clingo.Symbol objects.

    Example usage:
    symbol = json.loads(json, default=encoder)

    Args:
      obj: a JSON object
    '''
    if not isinstance(obj, Mapping): return obj
    if "clingo.SymbolType" not in obj: return obj
    stype_str = obj["clingo.SymbolType"]
    if stype_str == "Infimum": return clingo.Infimum
    if stype_str == "Supremum": return clingo.Supremum
    if stype_str == "String": return clingo.String(obj["string"])
    if stype_str == "Number": return clingo.Number(obj["number"])
    if stype_str == "Function":
        args = [ symbol_decoder(a) for a in obj["arguments"] ]
        positive = obj["positive"]
        return clingo.Function(obj["name"], args, positive)

    # A bad encoding?
    return obj

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------

class FactBaseCoder(object):
    '''A JSON Encoder/Decoder for clingo.Symbols, predicate instances, and fact bases.

    Provides a helper class for encoding and decoding facts to JSON. The
    predicates of interest are passed in the constructor or can be registered
    using a decorator.

    Args:
      predicates([Predicate]): a list of predicates to handle encoding/decoding

    '''
    def __init__(self, predicates=[]):
        self._preds = []
        self._predset = set()
        self._name2pred = {}
        for p in predicates: self._register_predicate(p)

    def _register_predicate(self, cls):
        if cls in self._predset: return    # ignore if already registered
        if not issubclass(cls, Predicate):
            raise TypeError("{} is not a Predicate sub-class".format(cls))
        self._predset.add(cls)
        self._preds.append(cls)
        self._name2pred[cls.__name__] = cls

    #-------------------------------------------------------------------------
    #
    #-------------------------------------------------------------------------
    def register(self, cls):
        '''Decorator to register a Predicate sub-class with the FactBaseCoder'''

        self._register_predicate(cls)
        return cls

    def encoder(self, obj):
        '''JSON Encoder for clingo.Symbol, clorm.Predicate, and clorm.FactBase.

        Call by overiding the ``default`` argument for json.dump(s)

        class Fun(Predicate):
           aint = IntegerField()

        fc = FactCoder(predicates=[Fun])
        return json.dumps([Fun(aint=1), Fun(aint2)], default=fc.encoder)

        Args:
          obj: an object to encode as json
        '''
        if isinstance(obj, clingo.Symbol): return symbol_encoder(obj)
        if isinstance(obj, FactBase):
            return {
                "clorm.FactBase" : [ str(fp) for fp in obj.indexes ],
                "facts" : [ self.encoder(fct) for fct in obj] }
        for p in self._preds:
            if isinstance(obj, p):
                return { "clorm.Predicate" : p.__name__,
                         "raw" : symbol_encoder(obj.raw) }
        _raise(obj)

    def decoder(self, obj):
        '''JSON Decoder for clingo.Symbol, clorm.Predicate, and clorm.FactBase.

        Call by overiding the ``object_hook`` argument for json.load(s)

        return json.dumps(obj, default=self.encoder)

        class Fun(Predicate):
           aint = IntegerField()

        fc = FactCoder(predicates=[Fun])
        return json.loads(json_str, object_hook=fc.encoder)

        Args:
          json_object: a json encoded object

        '''
        if not isinstance(obj, Mapping): return obj
        if "clingo.SymbolType" in obj: return symbol_decoder(obj)
        if "clorm.FactBase" in obj and "facts" in obj:
            indexes = []
            for fname in obj["clorm.FactBase"]:
                fs = fname.split('.')
                if len(fs) < 2:
                    raise ValueError(("Expecting a field '.' split for index "
                                      "{}").format(fs))
                if fs[0] not in self._name2pred:
                    raise ValueError(("Unrecognised predicate name {} not one "
                                      "of {}").format(fs, self._name2pred.keys()))
                ppath = path(self._name2pred[fs[0]])
                for key in fs[1:]: ppath = ppath[key]
                indexes.append(ppath)
            facts = [ self.decoder(f) for f in obj["facts"] ]
            return FactBase(facts=facts, indexes=indexes)
        if not "clorm.Predicate" in obj: return obj
        pname = obj["clorm.Predicate"]
        if pname not in self._name2pred: return obj
        return self._name2pred[pname](raw=symbol_decoder(obj["raw"]))

    #-------------------------------------------------------------------------
    # Convenience functions to call the JSON encoder and decoder
    #-------------------------------------------------------------------------

    def dumps(self, obj, indent=None, sort_keys=False):
        '''A convenience function for calling json.dumps'''
        return json.dumps(obj, indent=indent,
                          sort_keys=sort_keys, default=self.encoder)

    def dump(self, obj, fp, indent=None, sort_keys=False):
        '''A convenience function for calling json.dump'''
        return json.dump(obj, fp, indent=indent,
                         sort_keys=sort_keys, default=self.encoder)

    def loads(self, json_str):
        '''A convenience function for calling json.loads'''
        return json.loads(json_str, object_hook=self.decoder)

    def load(self, fp):
        '''A convenience function for calling json.load'''
        return json.load(fp, object_hook=self.decoder)


#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')



#--------------------------------------------------------------------------------
# A library to use instead of the clingo module when we want to convert to/from
# clingo symbol objects without actually using clingo symbol objects. The
# advantage is that it is just a basic python object and doesn't have the
# garbage collection issues that the real clingo symbol objects have. Useful for
# distributed applications.
# --------------------------------------------------------------------------------
import functools
import enum
import clingo

__all__ = [
    'Function',
    'String',
    'Number',
    'Infimum',
    'Supremum',
    'SymbolType',
    'is_Number',
    'is_Function',
    'is_String',
    'is_Infimum',
    'is_Supremum',
    'is_Symbol',
    'clingo_to_noclingo',
    'noclingo_to_clingo',
    'SymbolMode',
    'get_symbol_generator'
]

class SymbolType(enum.IntEnum):
    Infimum = 1
    Number = 2
    String = 3
    Function = 4
    Supremum = 5

    def __str__(self):
        if self.Infimum: return "Infimum"
        elif self.Number: return "Number"
        elif self.String: return "String"
        elif self.Function: return "Function"
        return "Supremum"


class Symbol(object):
    """A noclingo replacement for clingo.Symbol.

    noclingo is a replacement for clingo that provides basic Symbol object
    instantiation as pure python objects (rather than the C clingo external
    library). This allows a process to create a problem instance using the clorm
    predicate/complex term objects, JSON serialise the instance (using the
    clorm.json module) and pass the instance to a second process for actual
    processing.

    This is useful because clingo has a known issue that it cannot release
    Symbol objects. The solver was originally designed to solve one problem and
    exit. It has slowly evolved into calling the solver repeated as part of a
    larger application, but no facility has been added (yet) to allow the old
    Symbol objects to be released.
    """

    __slots__=("_stype","_args","_value","_sign","_hash")

    def __init__(self, stype, value=None, args=[],sign=True):
        if not isinstance(stype, SymbolType):
            raise TypeError("{} is not a SymbolType".format(stype))
        self._stype = stype
        self._args = None
        self._value = None
        self._sign = None
        if stype == SymbolType.Infimum:
            self._hash = hash(0)
        elif stype == SymbolType.Supremum:
            self._hash = hash(100)
        elif stype == SymbolType.Number:
            self._value = int(value)
            self._hash = hash(self._value)
        elif stype == SymbolType.String:
            self._value = str(value)
            self._hash = hash(self._value)
        elif stype == SymbolType.Function:
            self._sign = bool(sign)
            self._value = str(value)
            self._args = tuple(args)
            self._hash = hash((self._value,self._args,self._sign))

            if not self._value and not self._sign:
                raise ValueError("Tuple symbol cannot have a negative sign")
        else:
            raise ValueError("Unknown SymbolType {}".format(stype))

    @property
    def name(self):
        if self._stype != SymbolType.Function: return None
        return self._value

    @property
    def arguments(self):
        if self._stype != SymbolType.Function: return None
        return self._args

    @property
    def string(self):
        if self._stype != SymbolType.String: return None
        return self._value

    @property
    def number(self):
        if self._stype != SymbolType.Number: return None
        return self._value

    @property
    def type(self):
        return self._stype

    @property
    def positive(self):
        if self._stype != SymbolType.Function: return None
        return self._sign

    @property
    def negative(self):
        if self._stype != SymbolType.Function: return None
        return not self._sign

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        """Overloaded boolean operator."""
        if not isinstance(other, self.__class__): return NotImplemented
        if self._stype != other._stype: return False
        if self._stype == SymbolType.Infimum: return True
        if self._stype == SymbolType.Supremum: return True
        if self._stype != SymbolType.Function: return self._value == other._value

        # SymbolType.Function
        if self._hash != other._hash: return False
        if self._value != other._value: return False
        return self._args == other._args

    def __ne__(self, other):
        """Overloaded boolean operator."""
        result = self.__eq__(other)
        if result is NotImplemented: return NotImplemented
        return not result

    def __gt__(self, other):
        """Overloaded boolean operator."""
        if not isinstance(other, self.__class__): return NotImplemented
        if self._stype != other._stype: return self._stype > other._stype
        if self._stype == SymbolType.Infimum: return False
        if self._stype == SymbolType.Supremum: return False

        if self._stype != SymbolType.Function: return self._value > other._value
        if self._value != other._value: return self._value < other._value
        return self._args > other._args

    def __le__(self, other):
        """Overloaded boolean operator."""
        result = self.__gt__(other)
        if result is NotImplemented: return NotImplemented
        return not result

    def __lt__(self, other):
        """Overloaded boolean operator."""
        if not isinstance(other, self.__class__): return NotImplemented
        if self._stype != other._stype: return self._stype < other._stype
        if self._stype == SymbolType.Infimum: return False
        if self._stype == SymbolType.Supremum: return False

        if self._stype != SymbolType.Function: return self._value < other._value
        if self._value != other._value: return self._value < other._value
        return self._args < other._args

    def __ge__(self, other):
        """Overloaded boolean operator."""
        result = self.__lt__(other)
        if result is NotImplemented: return NotImplemented
        return not result

    def __str__(self):
        if self._stype == SymbolType.Number: return str(self._value)
        if self._stype == SymbolType.String: return '"' + self._value + '"'
        if self._stype == SymbolType.Infimum: return "#inf"
        if self._stype == SymbolType.Supremum: return "#sup"

        # SymbolType.Function
        if not self._args: return str(self._value)
        return "{}{}({})".format("" if self._sign else "-", self._value,
                                 ",".join([str(a) for a in self._args]))

    def __repr__(self):
        return self.__str__()

#--------------------------------------------------------------------------------
# Constants
#--------------------------------------------------------------------------------

Infimum = Symbol(SymbolType.Infimum)
Supremum = Symbol(SymbolType.Supremum)

#--------------------------------------------------------------------------------
# helper functions to create objects
#--------------------------------------------------------------------------------

def Function(name, args=[],sign=True):
    return Symbol(SymbolType.Function,name,args,sign)

def String(string):
    return Symbol(SymbolType.String,string)

def Number(number):
    return Symbol(SymbolType.Number,number)


#--------------------------------------------------------------------------------
# Functions to convert between clingo.Symbol and noclingo.Symbol
# --------------------------------------------------------------------------------

def clingo_to_noclingo(clsym):
    if clsym.type == clingo.SymbolType.Infimum: return Infimum
    elif clsym.type == clingo.SymbolType.Supremum: return Supremum
    elif clsym.type == clingo.SymbolType.Number: return Number(clsym.number)
    elif clsym.type == clingo.SymbolType.String: return String(clsym.string)
    elif clsym.type != clingo.SymbolType.Function:
        raise TypeError(("Symbol '{}' ({}) is not of type clingo.SymbolType."
                         "Function").format(clsym,type(clsym)))

    return Function(clsym.name,
                    (clingo_to_noclingo(t) for t in clsym.arguments),
                    clsym.positive)


def noclingo_to_clingo(nclsym):
    if nclsym.type == SymbolType.Infimum: return clingo.Infimum
    elif nclsym.type == SymbolType.Supremum: return clingo.Supremum
    elif nclsym.type == SymbolType.Number: return clingo.Number(nclsym.number)
    elif nclsym.type == SymbolType.String: return clingo.String(nclsym.string)
    elif nclsym.type != SymbolType.Function:
        raise TypeError(("Symbol '{}' ({}) is not of type noclingo.SymbolType."
                         "Function").format(nclsym,type(nclsym)))

    return clingo.Function(nclsym.name,
                    tuple(noclingo_to_clingo(t) for t in nclsym.arguments),
                    nclsym.positive)


# ------------------------------------------------------------------------------
# Internal function that converts the clingo.SymbolType to noclingo.SymbolType
# ------------------------------------------------------------------------------
def _get_symboltype(sym):
    if isinstance(sym,clingo.Symbol):
        if sym.type == clingo.SymbolType.Number: return SymbolType.Number
        elif sym.type == clingo.SymbolType.String: return SymbolType.String
        elif sym.type == clingo.SymbolType.Function: return SymbolType.Function
        elif sym.type == clingo.SymbolType.Infimum: return SymbolType.Infimum
        elif sym.type == clingo.SymbolType.Supremum: return SymbolType.Supremum
        else:
            raise ValueError(("Internal Error: unrecognised SymbolType for "
                              "'{}'").format(sym))
    if isinstance(sym, Symbol): return sym.type
    raise TypeError("Object '{}' ({}) is not a Symbol".format(sym,type(sym)))

# ------------------------------------------------------------------------------
# Functions to test the type of Symbol irrespective of clingo.Symbol or
# noclingo.Symbol.
# ------------------------------------------------------------------------------
def is_Number(sym):
    try:
        stype = _get_symboltype(sym)
        return stype == SymbolType.Number
    except:
        return False

def is_String(sym):
    try:
        stype = _get_symboltype(sym)
        return stype == SymbolType.String
    except:
        return False

def is_Function(sym):
    try:
        stype = _get_symboltype(sym)
        return stype == SymbolType.Function
    except:
        return False

def is_Supremum(sym):
    try:
        stype = _get_symboltype(sym)
        return stype == SymbolType.Supremum
    except:
        return False

def is_Infimum(sym):
    try:
        stype = _get_symboltype(sym)
        return stype == SymbolType.Infimum
    except:
        return False

def is_Symbol(sym):
    try:
        stype = _get_symboltype(sym)
        return True
    except:
        return False

#------------------------------------------------------------------------------
# A mechanism to group together the symbol generator functions for clingo or
# noclingo.
# ------------------------------------------------------------------------------

class SymbolMode(enum.IntEnum):
    CLINGO=0
    NOCLINGO=1

class SymbolGenerator(object):
    """Groups together the Symbol generators for clingo or noclingo.

    noclingo is a mirror of the clingo ``Symbol`` class that creates a Python
    only object. ``noclingo.Symbol`` objects can be used as a proxy for
    ``clingo.Symbol`` in places where you don't need to pass the object to the
    clingo solver. The advantage is that this object is not persistent
    throughout the life time of the process.

    The available member functions and properties are:

    * ``Infimum``

    * ``Supremum``

    * ``String()``

    * ``Number()``

    * ``Function()``
    """

    def __init__(self, mode, **kwargs):
        self._mode = mode
        self._links = dict(kwargs)

    @property
    def mode(self): return self._mode

    def __getattr__(self, item):
        return self._links[item]


clingo_symbol_generator = SymbolGenerator(SymbolMode.CLINGO,
                                          Function=clingo.Function,
                                          String=clingo.String,
                                          Number=clingo.Number,
                                          Infimum=clingo.Infimum,
                                          Supremum=clingo.Supremum,
                                          SymbolType=clingo.SymbolType)

noclingo_symbol_generator = SymbolGenerator(SymbolMode.NOCLINGO,
                                            Function=Function,
                                            String=String,
                                            Number=Number,
                                            Infimum=Infimum,
                                            Supremum=Supremum,
                                            SymbolType=SymbolType)

def get_symbol_generator(mode):
    if mode == SymbolMode.CLINGO: return clingo_symbol_generator
    if mode == SymbolMode.NOCLINGO: return noclingo_symbol_generator
    raise ValueError("Unknown SymbolMode {}".format(mode))


#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')



# -----------------------------------------------------------------------------
# Supporting containers for FactBase. In particular FactIndex, FactSet, and
# FactMap.
# ------------------------------------------------------------------------------

import io
import operator
import collections
import bisect
import inspect
import abc
import functools
import itertools

from .core import *
from .core import notcontains, PredicatePath, \
    get_field_definition, kwargs_check_keys

# ------------------------------------------------------------------------------
# In order to implement FactBase I originally used the built in 'set'
# class. However this uses the hash value, which for Predicate instances uses
# the underlying clingo.Symbol.__hash__() function. This in-turn depends on the
# c++ std::hash function. Like the Python standard hash function this uses
# random seeds at program startup which means that between successive runs of
# the same program the ordering of the set can change. This is bad for producing
# deterministic ASP solving. So using an OrderedSet instead.

from ..util import OrderedSet as FactSet

from ..util import OrderedSet

#FactSet=set                                # The Python standard set class. Note
                                           # fails some unit tests because I'm
                                           # testing for the ordering.

# Note: Some other 3rd party libraries that were tried but performed worse on
# a basic FactBase creation process:
#
#from ordered_set import OrderedSet as FactSet

#from orderedset import OrderedSet as FactSet   # Note: broken implementation so
                                               # fails some unit tests - union
                                               # operator only accepts a single
                                               # argument

#from blist import sortedset as FactSet
#from sortedcontainers import SortedSet as FactSet
# ------------------------------------------------------------------------------

__all__ = [
    'FactSet',
    'FactIndex',
    'FactMap',
    'factset_equality',
    ]

#------------------------------------------------------------------------------
# Global
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# FactIndex indexes facts by a given field
#------------------------------------------------------------------------------

class FactIndex(object):
    def __init__(self, path):
        try:
            self._path = path
            self._attrgetter = self._path.meta.attrgetter
            self._predicate = self._path.meta.predicate
            self._keylist = []
            self._key2values = collections.OrderedDict()
        except:
            raise TypeError("{} is not a valid PredicatePath object".format(path))

    @property
    def path(self):
        return self._path

    def add(self, fact):
        if not isinstance(fact, self._predicate):
            raise TypeError("{} is not a {}".format(fact, self._predicate))
        key = self._attrgetter(fact)

        # Index the fact by the key - Note: using OrderedSet to preserve
        # insertion order for repeatability
        if key not in self._key2values: self._key2values[key] = OrderedSet()
        self._key2values[key].add(fact)

        # Maintain the sorted list of keys
        posn = bisect.bisect_left(self._keylist, key)
        if len(self._keylist) > posn and self._keylist[posn] == key: return
        bisect.insort_left(self._keylist, key)

    def discard(self, fact):
        self.remove(fact, False)

    def remove(self, fact, raise_on_missing=True):
        if not isinstance(fact, self._predicate):
            raise TypeError("{} is not a {}".format(fact, self._predicate))
        key = self._attrgetter(fact)

        # Remove the value
        if key not in self._key2values:
            if raise_on_missing:
                raise KeyError("{} is not in the FactIndex".format(fact))
            return
        values = self._key2values[key]
        if raise_on_missing: values.remove(fact)
        else: values.discard(fact)

        # If still have values then we're done
        if values: return

        # remove the key
        del self._key2values[key]
        posn = bisect.bisect_left(self._keylist, key)
        del self._keylist[posn]

    def clear(self):
        self._keylist = []
        self._key2values = collections.OrderedDict()

    @property
    def keys(self): return self._keylist

    #--------------------------------------------------------------------------
    # Internal functions to get keys matching some boolean operator
    #--------------------------------------------------------------------------

    def _keys_eq(self, key):
        if key in self._key2values: return [key]
        return []

    def _keys_ne(self, key):
        posn1 = bisect.bisect_left(self._keylist, key)
        left = self._keylist[:posn1]
        posn2 = bisect.bisect_right(self._keylist, key)
        right = self._keylist[posn2:]
        return left + right

    def _keys_lt(self, key):
        posn = bisect.bisect_left(self._keylist, key)
        return self._keylist[:posn]

    def _keys_le(self, key):
        posn = bisect.bisect_right(self._keylist, key)
        return self._keylist[:posn]

    def _keys_gt(self, key):
        posn = bisect.bisect_right(self._keylist, key)
        return self._keylist[posn:]

    def _keys_ge(self, key):
        posn = bisect.bisect_left(self._keylist, key)
        return self._keylist[posn:]

    def _keys_contains(self, seq):
        tmp = []
        for key in seq:
            if key in self._key2values: tmp.append(key)
        tmp.sort()
        return tmp

    def _keys_notcontains(self, seq):
        tmp = []
        for key in self._keylist:
            if key not in seq:
                tmp.append(key)
        return tmp

    #--------------------------------------------------------------------------
    # Find elements based on boolean match to a key
    #--------------------------------------------------------------------------
    def find(self, op, val,reverse=False):
        keys = []
        if op == operator.eq: keys = self._keys_eq(val)
        elif op == operator.ne: keys = self._keys_ne(val)
        elif op == operator.lt: keys = self._keys_lt(val)
        elif op == operator.le: keys = self._keys_le(val)
        elif op == operator.gt: keys = self._keys_gt(val)
        elif op == operator.ge: keys = self._keys_ge(val)
        elif op == operator.contains: keys = self._keys_contains(val)
        elif op == notcontains: keys = self._keys_notcontains(val)
        else: raise ValueError("unsupported operator {}".format(op))

        if reverse:
            for k in reversed(keys):
                for fact in self._key2values[k]: yield fact
        else:
            for k in keys:
                for fact in self._key2values[k]: yield fact

    #--------------------------------------------------------------------------
    # Iterate in descending key order
    #--------------------------------------------------------------------------

    def __reversed__(self):
        for key in reversed(self._keylist):
            for f in self._key2values[key]: yield f

    #--------------------------------------------------------------------------
    # Iterate in key ascending order
    #--------------------------------------------------------------------------

    def __iter__(self):
        for key in self._keylist:
            for f in self._key2values[key]: yield f

    def __bool__(self):
        for facts in self._key2values.values():
            if facts: return True
        return False

    def __len__(self):
        return sum([len(facts) for facts in self._key2values.values()])

    def __eq__(self, other):
        if not isinstance(other, self.__class__): return NotImplemented
        if hashable_path(self._path) != hashable_path(other._path): return False
        return self._key2values == other._key2values

    def __ne__(self, other):
        """Overloaded boolean operator."""
        result = self.__eq__(other)
        if result is NotImplemented: return NotImplemented
        return not result

    def __str__(self):
        if not self: return "{}"
        tmp = []
        for k,v in self._key2values.items(): tmp.extend(v)
        return "{" + ", ".join([repr(f) for f in tmp]) + "}"

    def __repr__(self):
        return self.__str__()

#------------------------------------------------------------------------------
# A helper function to determine if two collections have the same elements
# (irrespective of ordering). This is useful if the underlying objects are two
# OrderedSet objects since the equality operator will also test for the same
# ordering which is something we don't want.
# ------------------------------------------------------------------------------

def factset_equality(s1,s2):
    if len(s1) != len(s2): return False
    for elem in s1:
        if elem not in s2: return False
    return True

#------------------------------------------------------------------------------
# FactMap is simply a meta-container for FactSet and FactIndex containers.
#
# For each predicate type it holds a FactSet and any associated FactIndexes.
#
# FactMap only implements the operations for adding and deleting facts (and set
# operations that add/delete facts). The reason is that this is complicated by
# having to add and delete the facts from multiple source (the FactSet and
# multiple FactIndex objects).
#
# For accessing facts and testing the inclusion of a fact in the map you must
# access the underlying FactSet or FactIndex object. For example, the query
# engine is passed the FactMap and then choses the appropriate way of accessing
# the data.
#
# ------------------------------------------------------------------------------

def _fm_iterable(other):
    if isinstance(other, FactMap): return other.factset
    else: return other

class FactMap(object):
    def __init__(self, ptype, indexes=[]):
        def clean_path(p):
            p = path(p)
            if hashable_path(p) != hashable_path(p.meta.dealiased):
                raise ValueError(("It doesn't make sense to index on an alias "
                                  "'{}'").format(p))
            if p.meta.is_root:
                raise ValueError(("It doesn't make sense to index a root path "
                                  "'{}'").format(p))
            if p.meta.predicate != ptype:
                raise ValueError(("Cannot index path '{}' that doesn't match "
                                  "the predicate '{}'").format(p,ptype))
            return hashable_path(p)

        self._ptype = ptype
        self._factset = FactSet()
        self._path2factindex = collections.OrderedDict()
        self._factindexes = []

        # Validate the paths to be indexed
        allindexes = set([clean_path(p) for p in indexes])

        for pth in allindexes:
            tmppath=path(pth)
            if hashable_path(tmppath.meta.dealiased) != hashable_path(tmppath):
                raise ValueError(("Cannot create an index for an alias path "
                                  "'{}'").format(tmppath))
            if tmppath.meta.predicate != ptype:
                raise ValueError(("Index path '{}' isn't a sub-path of Predicate "
                                  "'{}'").format(tmppath, path(ptype)))
            tmpfi = FactIndex(tmppath)
            self._path2factindex[hashable_path(tmppath)] = tmpfi
            self._factindexes.append(tmpfi)
        self._factindexes = tuple(self._factindexes)

    def add_facts(self, facts):
        for f in facts:
            self._factset.add(f)
            for fi in self._factindexes: fi.add(f)

    def add_fact(self, fact):
        self._factset.add(fact)
        for fi in self._factindexes: fi.add(fact)

    def discard(self,fact):
        self.remove(fact,False)

    def remove(self,fact, raise_on_missing=True):
        if raise_on_missing: self._factset.remove(fact)
        else: self._factset.discard(fact)
        for fi in self._factindexes:
            fi.remove(fact,raise_on_missing)

    def pop(self):
        if not self._factset: raise KeyError("Cannot pop() an empty set of facts")
        fact = next(iter(self._factset))
        self.remove(fact)
        return fact

    def clear(self):
        self._factset.clear()
        for fi in self._factindexes: fi.clear()

    @property
    def predicate(self):
        return self._ptype

    @property
    def factset(self):
        return self._factset

    @property
    def path2factindex(self):
        return self._path2factindex

    def __bool__(self):
        return bool(self._factset)

    #--------------------------------------------------------------------------
    # Set functions
    #--------------------------------------------------------------------------
    def union(self,*others):
        nfm = FactMap(self.predicate, self._path2factindex.keys())
        tmpothers = [_fm_iterable(o) for o in others]
        tmp = self.factset.union(*tmpothers)
        nfm.add_facts(tmp)
        return nfm

    def intersection(self,*others):
        nfm = FactMap(self.predicate, self._path2factindex.keys())
        tmpothers = [_fm_iterable(o) for o in others]
        tmp = self.factset.intersection(*tmpothers)
        nfm.add_facts(tmp)
        return nfm

    def difference(self,*others):
        nfm = FactMap(self.predicate, self._path2factindex.keys())
        tmpothers = [_fm_iterable(o) for o in others]
        tmp = self.factset.difference(*tmpothers)
        nfm.add_facts(tmp)
        return nfm

    def symmetric_difference(self,other):
        nfm = FactMap(self.predicate, self._path2factindex.keys())
        tmp = self.factset.symmetric_difference(_fm_iterable(other))
        nfm.add_facts(tmp)
        return nfm

    def update(self,*others):
        self.add_facts(itertools.chain(*[_fm_iterable(o) for o in others]))

    def intersection_update(self,*others):
        for f in set(self.factset):
            for o in others:
                if f not in _fm_iterable(o): self.discard(f)

    def difference_update(self,*others):
        for f in itertools.chain(*[o.factset for o in others]):
            self.discard(f)

    def symmetric_difference_update(self, other):
        to_remove=OrderedSet()
        to_add=OrderedSet()
        for f in self._factset:
            if f in _fm_iterable(other): to_remove.add(f)
        for f in _fm_iterable(other):
            if f not in self._factset: to_add.add(f)
        for f in to_remove: self.discard(f)
        self.add_facts(to_add)

    def copy(self):
        nfm = FactMap(self.predicate, self._path2factindex.keys())
        nfm.add_facts(self.factset)
        return nfm


#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')



# -----------------------------------------------------------------------------
# Clorm ORM FactBase implementation. FactBase provides a set-like container
# specifically for storing facts (Predicate instances).
# ------------------------------------------------------------------------------

import sys
import io
import operator
import collections
import bisect
import inspect
import abc
import functools
import itertools

from .core import *
from .core import get_field_definition, PredicatePath, kwargs_check_keys, \
    validate_root_paths

from .query import *

from .query import Placeholder, OrderBy, desc, asc

from .query import process_where, process_join, process_orderby, \
    make_query_plan, QuerySpec, QueryExecutor

from .factcontainers import FactSet, FactIndex, FactMap, factset_equality

__all__ = [
    'FactBase',
    'Select',
    'Delete',
    ]

#------------------------------------------------------------------------------
# Global
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Support function for printing ASP facts: Note: _trim_docstring() is taken from
# PEP 257 (modified for Python 3): https://www.python.org/dev/peps/pep-0257/
# ------------------------------------------------------------------------------

def _format_asp_facts(iterator,output,width):
    tmp1=""
    for f in iterator:
        fstr="{}.".format(f)
        if tmp1 and len(tmp1) + len(fstr) > width:
            print(tmp1,file=output)
            tmp1 = fstr
        else:
            tmp1 = tmp1 + " " + fstr if tmp1 else fstr
    if tmp1: print(tmp1,file=output)

def _trim_docstring(docstring):
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)

def _endstrip(string):
    if not string: return
    nl=string[-1]=='\n'
    tmp=string.rstrip()
    return tmp + '\n' if nl else tmp

def _format_docstring(docstring,output):
    if not docstring: return
    tmp=_trim_docstring(docstring)
    tmpstr = "".join(_endstrip("%     " + l) for l in tmp.splitlines(True))
    if tmpstr:
        print("% Description:",file=output)
        print(tmpstr,file=output)

def _maxwidth(lines):
    return max([len(l) for l in lines])

def _format_commented(fm,out):
    pm = fm.predicate.meta
    docstring = _trim_docstring(fm.predicate.__doc__) \
        if fm.predicate.__doc__ else ""
    indent = "    "
    if pm.arity == 0:
        lines = [ "Unary predicate signature:", indent + pm.name ]
    else:
        params=[str(fa.name) for fa in pm]
        lines = [ "Predicate signature:",
                  indent + "{}({})".format(pm.name, ",".join(params)) ]
    if docstring:
        lines.append("Description:")
        for l in docstring.splitlines():lines.append(indent + l)
    bar = "-" * _maxwidth(lines)
    lines.insert(0,bar)
    lines.append(bar)
    for l in lines:
        tmp = l.rstrip()
        if tmp: print("% {}".format(tmp),file=out)
        else: print("%",file=out)
    return

#------------------------------------------------------------------------------
# A FactBase consisting of facts of different types
#------------------------------------------------------------------------------

class FactBase(object):
    """A fact base is a container for facts (i.e., Predicate sub-class instances)

    ``FactBase`` can be behave like a specialised ``set`` object, but can also
    behave like a minimalist database. It stores facts for ``Predicate`` types
    (where a predicate type loosely corresponds to a *table* in a database)
    and allows for certain fields to be indexed in order to perform more
    efficient queries.

    The initaliser can be given a collection of predicates. If it is passed
    another FactBase then it simply makes a copy (including the indexed fields).

    FactBase also has a special mode when it is passed a functor instead of a
    collection. In this case it performs a delayed initialisation. This means
    that the internal data structures are only populated when the FactBase is
    actually used. This mode is particularly useful when extracting facts from
    models. Often a program will only want to keep the data from the final model
    (for example, with optimisation we often want the best model before a
    timeout). Delayed initialisation is useful will save computation as only the
    last model will be properly initialised.

    Args:
      facts([Predicate]|FactBase|callable): a list of facts (predicate
         instances), a fact base, or a functor that generates a list of
         facts. If a functor is passed then the fact base performs a delayed
         initialisation. If a fact base is passed and no index is specified then
         an index will be created matching in input fact base.
      indexes(Field): a list of fields that are to be indexed.

    """

    #--------------------------------------------------------------------------
    # Internal member functions
    #--------------------------------------------------------------------------

    # A special purpose initialiser so that we can delayed initialisation
    def _init(self, facts=None, indexes=None):

        # flag that initialisation has taken place
        self._delayed_init = None

        # If it is delayed initialisation then get the facts
        if facts and callable(facts):
            facts = facts()
        elif facts and isinstance(facts, FactBase) and indexes is None:
            indexes = facts.indexes
        if indexes is None: indexes=[]

        # Create FactMaps for the predicate types with indexed fields
        grouped = {}

        self._indexes = tuple(indexes)
        for path in self._indexes:
            if path.meta.predicate not in grouped: grouped[path.meta.predicate] = []
            grouped[path.meta.predicate].append(path)
        self._factmaps = { pt : FactMap(pt, idxs) for pt, idxs in grouped.items() }

        if facts is None: return
        self._add(facts)

    # Make sure the FactBase has been initialised
    def _check_init(self):
        if self._delayed_init: self._delayed_init()  # Check for delayed init

    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------

    def _add(self, arg):
        if isinstance(arg, Predicate):
            return self._add_fact(type(arg),arg)
        elif isinstance(arg, str) or not isinstance(arg, collections.abc.Iterable):
            raise TypeError(f"'{arg}' is not a Predicate instance")

        facts = sorted(arg, key=lambda x : type(x).__name__)
        for ptype, g in itertools.groupby(facts, lambda x: type(x)):
            self._add_facts(ptype, g)

    def _add_fact(self, ptype, fact):
        fm = self._factmaps.setdefault(ptype, FactMap(ptype))
        fm.add_fact(fact)

    def _add_facts(self, ptype, facts):
        if not issubclass(ptype,Predicate):
            raise TypeError(f"{list(facts)} are not Predicate instances")
        fm = self._factmaps.setdefault(ptype, FactMap(ptype))
        fm.add_facts(facts)

    def _remove(self, fact, raise_on_missing):
        ptype = type(fact)
        if not isinstance(fact, Predicate) or ptype not in self._factmaps:
            raise KeyError(fact)

        return self._factmaps[ptype].remove(fact,raise_on_missing)

    #--------------------------------------------------------------------------
    # Initiliser
    #--------------------------------------------------------------------------
    def __init__(self, facts=None, indexes=None):
        self._delayed_init=None
        if callable(facts):
            def delayed_init():
                self._init(facts, indexes)
            self._delayed_init=delayed_init
        else:
            self._init(facts, indexes)


    #--------------------------------------------------------------------------
    # An internal API for the query mechanism. Not to be called by users.
    #--------------------------------------------------------------------------
    @property
    def factmaps(self):
        self._check_init()  # Check for delayed init
        return self._factmaps

    #--------------------------------------------------------------------------
    # Set member functions
    #--------------------------------------------------------------------------
    def add(self, arg):
        """Add a single fact or a collection of facts.

        Because a ``FactBase`` can only hold :class:`~Predicate` sub-class
        instances this member function has been overloaded to take either a
        single :class:`~Predicate` sub-class instance or a collection of
        :class:`~Predicate` sub-class instances.

        Args:
          arg: a single fact or a collection of facts.

        """
        self._check_init()  # Check for delayed init
        return self._add(arg)

    def remove(self, arg):
        """Remove a fact from the fact base (raises an exception if no fact). """
        self._check_init()  # Check for delayed init
        return self._remove(arg, raise_on_missing=True)

    def discard(self, arg):
        """Remove a fact from the fact base. """
        self._check_init()  # Check for delayed init
        return self._remove(arg, raise_on_missing=False)

    def pop(self):
        """Pop an element from the FactBase. """
        self._check_init()  # Check for delayed init
        for pt, fm in self._factmaps.items():
            if fm: return fm.pop()
        raise KeyError("pop from an empty FactBase")

    def clear(self):
        """Clear the fact base of all facts."""

        self._check_init()  # Check for delayed init
        for pt, fm in self._factmaps.items(): fm.clear()

    #--------------------------------------------------------------------------
    # Special FactBase member functions
    #--------------------------------------------------------------------------
    def select(self, root):
        """Define a select query using the old Query API.

        .. note::

           This interface will eventually be deprecated when the new
           :class:`Query API<Query>` is finalised. The entry point to this Query
           API is through the :meth:`FactBase.query` method.

        Args:
           predicate: The predicate to query.

        Returns:
           Returns a Select query object for specifying a query.

        """
        self._check_init()  # Check for delayed init

        roots = validate_root_paths([root])
        ptypes = set([ root.meta.predicate for root in roots])

        # Make sure there are factmaps for each referenced predicate type
        for ptype in ptypes: self._factmaps.setdefault(ptype, FactMap(ptype))

        return SelectImpl(self, QuerySpec(roots=roots))

    def delete(self, root):
        self._check_init()  # Check for delayed init

        roots = validate_root_paths([root])
        ptypes = set([ root.meta.predicate for root in roots])

        # Make sure there are factmaps for each referenced predicate type
        for ptype in ptypes: self._factmaps.setdefault(ptype, FactMap(ptype))

        return _Delete(self, QuerySpec(roots=roots))

    def query(self, *roots):
        """Define a query using the new Query API :class:`Query`.

        The parameters consist of a predicates (or aliases) to query (like an
        SQL FROM clause).

        Args:
           *predicates: predicate or predicate aliases

        Returns:
           Returns a Query object for specifying a query.

        """

        self._check_init()  # Check for delayed init

        # Make sure there are factmaps for each referenced predicate type
        ptypes = set([r.meta.predicate for r in validate_root_paths(roots)])
        for ptype in ptypes: self._factmaps.setdefault(ptype, FactMap(ptype))

        qspec = QuerySpec(roots=roots)
        return QueryImpl(self._factmaps, qspec)

    @property
    def predicates(self):
        """Return the list of predicate types that this fact base contains."""

        self._check_init()  # Check for delayed init
        return tuple([pt for pt, fm in self._factmaps.items() if fm])

    @property
    def indexes(self):
        self._check_init()  # Check for delayed init
        return self._indexes

    def facts(self):
        """Return all facts."""

        self._check_init()  # Check for delayed init
        tmp = [ fm.factset for fm in self._factmaps.values() if fm]
        return list(itertools.chain(*tmp))

    def asp_str(self,width=0,commented=False):
        """Return a string representation of the fact base that is suitable for adding
        to an ASP program

        """
        self._check_init()  # Check for delayed init
        out = io.StringIO()

        if not commented:
            _format_asp_facts(self,out,width)
        else:
            first=True
            for fm in self._factmaps.values():
                if first: first=False
                else: print("",file=out)
                _format_commented(fm,out)
                _format_asp_facts(fm.factset,out,width)

        data = out.getvalue()
        out.close()
        return data



    def __str__(self):
        self._check_init()  # Check for delayed init
        tmp = ", ".join([str(f) for f in self])
        return '{' + tmp + '}'

    def __repr__(self):
        return self.__str__()

    #--------------------------------------------------------------------------
    # Special functions to support set and container operations
    #--------------------------------------------------------------------------

    def __contains__(self, fact):
        """Implemement set 'in' operator."""

        self._check_init() # Check for delayed init

        if not isinstance(fact,Predicate): return False
        ptype = type(fact)
        if ptype not in self._factmaps: return False
        return fact in self._factmaps[ptype].factset

    def __bool__(self):
        """Implemement set bool operator."""

        self._check_init() # Check for delayed init

        for fm in self._factmaps.values():
            if fm: return True
        return False

    def __len__(self):
        self._check_init() # Check for delayed init
        return sum([len(fm.factset) for fm in self._factmaps.values()])

    def __iter__(self):
        self._check_init() # Check for delayed init

        for fm in self._factmaps.values():
            for f in fm.factset: yield f

    def __eq__(self, other):
        """Overloaded boolean operator."""

        # If other is not a FactBase then create one
        if not isinstance(other, self.__class__): other=FactBase(other)
        self._check_init(); other._check_init() # Check for delayed init

        self_fms = { p: fm for p,fm in self._factmaps.items() if fm }
        other_fms = { p: fm for p,fm in other._factmaps.items() if fm }
        if self_fms.keys() != other_fms.keys(): return False

        for p, fm1 in self_fms.items():
            fm2 = other_fms[p]
            if not factset_equality(fm1.factset,fm2.factset): return False

        return True

    def __ne__(self, other):
        """Overloaded boolean operator."""
        result = self.__eq__(other)
        if result is NotImplemented: return NotImplemented
        return not result

    def __lt__(self,other):
        """Implemement set < operator."""

        # If other is not a FactBase then create one
        if not isinstance(other, self.__class__): other=FactBase(other)
        self._check_init() ; other._check_init() # Check for delayed init

        self_fms = { p: fm for p,fm in self._factmaps.items() if fm }
        other_fms = { p: fm for p,fm in other._factmaps.items() if fm }
        if len(self_fms) > len(other_fms): return False

        known_ne=False
        for p, spfm in self_fms.items():
            if p not in other_fms: return False
            opfm = other_fms[p]
            if spfm.factset < opfm.factset: known_ne=True
            elif spfm.factset > opfm.factset: return False

        if known_ne: return True
        return False

    def __le__(self,other):
        """Implemement set <= operator."""

        if not isinstance(other, self.__class__): other=FactBase(other)
        self._check_init() ; other._check_init() # Check for delayed init

        self_fms = { p: fm for p,fm in self._factmaps.items() if fm }
        other_fms = { p: fm for p,fm in other._factmaps.items() if fm }
        if len(self_fms) > len(other_fms): return False

        for p, spfm in self_fms.items():
            if p not in other_fms: return False
            opfm = other_fms[p]
            if spfm.factset > opfm.factset: return False
        return True

    def __gt__(self,other):
        """Implemement set > operator."""
        if not isinstance(other, self.__class__): other=FactBase(other)
        return other.__lt__(self)

    def __ge__(self,other):
        """Implemement set >= operator."""
        if not isinstance(other, self.__class__): other=FactBase(other)
        return other.__le__(self)

    def __or__(self,other):
        """Implemement set | operator."""
        return self.union(other)

    def __and__(self,other):
        """Implemement set & operator."""
        return self.intersection(other)

    def __sub__(self,other):
        """Implemement set - operator."""
        return self.difference(other)

    def __xor__(self,other):
        """Implemement set ^ operator."""
        return self.symmetric_difference(other)

    def __ior__(self,other):
        """Implemement set |= operator."""
        self.update(other)
        return self

    def __iand__(self,other):
        """Implemement set &= operator."""
        self.intersection_update(other)
        return self

    def __isub__(self,other):
        """Implemement set -= operator."""
        self.difference_update(other)
        return self

    def __ixor__(self,other):
        """Implemement set ^= operator."""
        self.symmetric_difference_update(other)
        return self

    def __getstate__(self):
        self._check_init()
        return self.__dict__


    #--------------------------------------------------------------------------
    # Set functions
    #--------------------------------------------------------------------------
    def union(self,*others):
        """Implements the set union() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()

        fb=FactBase()
        predicates = set(self._factmaps.keys())
        for o in others: predicates.update(o._factmaps.keys())

        for p in predicates:
            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
            if p in self._factmaps:
                fb._factmaps[p] = self._factmaps[p].union(*pothers)
            else:
                fb._factmaps[p] = FactMap(p).union(*pothers)
        return fb

    def intersection(self,*others):
        """Implements the set intersection() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()

        fb=FactBase()
        predicates = set(self._factmaps.keys())
        for o in others: predicates.intersection_update(o._factmaps.keys())

        for p in predicates:
            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
            fb._factmaps[p] = self._factmaps[p].intersection(*pothers)
        return fb

    def difference(self,*others):
        """Implements the set difference() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()

        fb=FactBase()
        predicates = set(self._factmaps.keys())

        for p in predicates:
            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
            fb._factmaps[p] = self._factmaps[p].difference(*pothers)
        return fb

    def symmetric_difference(self,other):
        """Implements the set symmetric_difference() function"""
        if not isinstance(other, self.__class__): other=FactBase(other)
        self._check_init() # Check for delayed init
        other._check_init()

        fb=FactBase()
        predicates = set(self._factmaps.keys())
        predicates.update(other._factmaps.keys())

        for p in predicates:
            in_self = p in self._factmaps ; in_other = p in other._factmaps
            if in_self and in_other:
                fb._factmaps[p] = self._factmaps[p].symmetric_difference(other._factmaps[p])
            elif in_self:
                fb._factmaps[p] = self._factmaps[p].copy()
            elif in_other:
                fb._factmaps[p] = other._factmaps[p].copy()

        return fb

    def update(self,*others):
        """Implements the set update() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()

        for o in others:
            for p,fm in o._factmaps.items():
                if p in self._factmaps: self._factmaps[p].update(fm)
                else: self._factmaps[p] = fm.copy()

    def intersection_update(self,*others):
        """Implements the set intersection_update() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()
        num = len(others)

        predicates = set(self._factmaps.keys())
        for o in others: predicates.intersection_update(o._factmaps.keys())
        pred_to_delete = set(self._factmaps.keys()) - predicates

        for p in pred_to_delete: self._factmaps[p].clear()
        for p in predicates:
            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
            self._factmaps[p].intersection_update(*pothers)

    def difference_update(self,*others):
        """Implements the set difference_update() function"""
        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
        self._check_init() # Check for delayed init
        for o in others: o._check_init()

        for p in self._factmaps.keys():
            pothers=[ o._factmaps[p] for o in others if p in o._factmaps ]
            self._factmaps[p].difference_update(*pothers)

    def symmetric_difference_update(self,other):
        """Implements the set symmetric_difference_update() function"""
        if not isinstance(other, self.__class__): other=FactBase(other)
        self._check_init() # Check for delayed init
        other._check_init()

        predicates = set(self._factmaps.keys())
        predicates.update(other._factmaps.keys())

        for p in predicates:
            if p in self._factmaps and p in other._factmaps:
                self._factmaps[p].symmetric_difference_update(other._factmaps[p])
            else:
                if p in other._factmaps: self._factmaps[p] = other._factmaps[p].copy()


    def copy(self):
        """Implements the set copy() function"""
        self._check_init() # Check for delayed init
        fb=FactBase()
        for p,fm in self._factmaps.items():
            fb._factmaps[p] = self._factmaps[p].copy()
        return fb

#------------------------------------------------------------------------------
# Select is an interface query over a FactBase.
# ------------------------------------------------------------------------------

class Select(abc.ABC):
    """An abstract class that defines the interface to original Query API.

    .. note::

       This interface will eventually be deprecated when the new :class:`Query
       API<Query>` is finalised.

    ``Select`` query objects cannot be constructed directly. Instead a
    ``Select`` object is returned by the :meth:`FactBase.select` function. Given
    a ``FactBase`` object ``fb``, a specification is of the form:

          ``query = fb.select(<predicate>).where(<expression>).order_by(<ordering>)``

    where ``<predicate>`` specifies the predicate type to search for,
    ``<expression>`` specifies the search criteria and ``<ordering>`` specifies
    a sort order when returning the results. The ``where()`` and ``order_by()``
    clauses are omitted when not required.

    """

    @abc.abstractmethod
    def where(self, *expressions):
        """Set the select statement's where clause.

        The where clause consists of a set of boolean and comparison
        expressions. This expression specifies a search criteria for matching
        facts within the corresponding ``FactBase``.

        Boolean expression are built from other boolean expression or a
        comparison expression. Comparison expressions are of the form:

               ``<PredicatePath> <compop>  <value>``

       where ``<compop>`` is a comparison operator such as ``==``, ``!=``, or
       ``<=`` and ``<value>`` is either a Python value or another predicate path
       object refering to a field of the same predicate or a placeholder.

        A placeholder is a special value that issubstituted when the query is
        actually executed. These placeholders are named ``ph1_``, ``ph2_``,
        ``ph3_``, and ``ph4_`` and correspond to the 1st to 4th arguments of the
        ``get``, ``get_unique`` or ``count`` function call.

        Args:
          expressions: one or more comparison expressions.

        Returns:
          Returns a reference to itself.

        """

    @abc.abstractmethod
    def order_by(self, *fieldorder):
        """Provide an ordering over the results.

        Args:
          fieldorder: an ordering over fields
        Returns:
          Returns a reference to itself.
        """
        pass

    @abc.abstractmethod
    def get(self, *args, **kwargs):
        """Return all matching entries."""
        pass

    def get_unique(self, *args, **kwargs):
        """Return the unique matching entry (or raise an exception)"""
        pass

    def count(self, *args, **kwargs):
        """Return the number of matches."""
        pass

#------------------------------------------------------------------------------
# Delete is an interface to perform a query delete from a FactBase.
# ------------------------------------------------------------------------------

class Delete(abc.ABC):
    """An abstract class that defines the interface to a original delete query API.

    .. note::

       This interface will eventually be deprecated when the new :class:`Query
       API<Query>` is finalised.

    ``Delete`` query objects cannot be constructed directly. Instead a
    ``Delete`` object is returned by the ``FactBase.delete()`` function. Given a
    ``FactBase`` object ``fb``, a specification is of the form:

          ``query = fb.delete(<predicate>).where(<expression>)``

    where ``<predicate>`` specifies the predicate type to search for,
    ``<expression>`` specifies the search criteria. The ``where()`` clause can
    be omitted in which case all predicates of that type will be deleted.

    """

    @abc.abstractmethod
    def where(self, *expressions):
        """Set the select statement's where clause.

        See the documentation for ``Select.where()`` for further details.
        """
        pass

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Function to execute the delete query"""
        pass

#------------------------------------------------------------------------------
# Query API version 1 with new query engine
#------------------------------------------------------------------------------

class SelectImpl(Select):

    def __init__(self, factbase, qspec):
        self._factbase = factbase
        self._qspec = qspec

    #--------------------------------------------------------------------------
    # Add an order_by expression
    #--------------------------------------------------------------------------
    def where(self, *expressions):
        if self._qspec.where:
            raise TypeError("Cannot specify 'where' multiple times")
        if not expressions:
            raise TypeError("Empty 'where' expression")

        try:
            if len(expressions) == 1:
                where = process_where(expressions[0],self._qspec.roots)
            else:
                where = process_where(and_(*expressions),self._qspec.roots)
            nqspec = self._qspec.newp(where=where)
        except ValueError as e:
            raise TypeError(str(e)) from None
        return SelectImpl(self._factbase,nqspec)

    #--------------------------------------------------------------------------
    # Add an order_by expression
    #--------------------------------------------------------------------------
    def order_by(self, *expressions):
        if self._qspec.order_by:
            raise TypeError("Cannot specify 'order_by' multiple times")
        if not expressions:
            raise TypeError("Empty 'order_by' expression")
        try:
            order_by=process_orderby(expressions,self._qspec.roots)
            nqspec = self._qspec.newp(order_by=order_by)
        except ValueError as e:
            raise TypeError(str(e)) from None
        return SelectImpl(self._factbase,nqspec)

    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    def query_plan(self,*args,**kwargs):
        qspec = self._qspec.fill_defaults()

        (factsets,factindexes) = \
            QueryExecutor.get_factmap_data(self._factbase.factmaps, qspec)
        qplan = make_query_plan(factindexes.keys(), qspec)

        return qplan.ground(*args,**kwargs)

    #--------------------------------------------------------------------------
    # Functions currently mirroring the old interface
    # --------------------------------------------------------------------------

    def get(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        return list(qe.all())

    def get_unique(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        found = None
        for out in qe.all():
            if found: raise ValueError("Query returned more than a single element")
            found = out
        return found

    def count(self, *args, **kwargs):
        qspec = self._qspec
        if args or kwargs:
            if self._qspec.where is None:
                raise ValueError(("No where clause to ground"))
            qspec = self._qspec.bindp(*args, **kwargs)

        qe = QueryExecutor(self._factbase.factmaps, qspec)
        count = 0
        for _ in qe.all(): count += 1
        return count

#------------------------------------------------------------------------------
# The Delete class
#------------------------------------------------------------------------------

class _Delete(Delete):

    def __init__(self, factbase, qspec):
        self._factbase = factbase
        self._root = qspec.roots[0]
        self._select = SelectImpl(factbase,qspec)
        self._has_where = False

    def where(self, *expressions):
        self._has_where = True
        self._select = self._select.where(*expressions)
        return self

    def execute(self, *args, **kwargs):
        factmap = self._factbase.factmaps[self._root.meta.predicate]

        # If there is no where clause then delete everything
        if not self._has_where:
            num_deleted = len(factmap.facts())
            factmap.clear()
            return num_deleted

        # Gather all the facts to delete and remove them
        to_delete = [ f for f in self._select.get(*args, **kwargs) ]
        for fact in to_delete: factmap.remove(fact)
        return len(to_delete)


#------------------------------------------------------------------------------
# New Clorm Query API
#
# QueryImpl
# - factmaps             - dictionary mapping predicate types to FactMap objects
# - qspec                - a dictionary with query parameters
#------------------------------------------------------------------------------
class QueryImpl(Query):

    def __init__(self, factmaps, qspec):
        self._factmaps = factmaps
        self._qspec = qspec

    #--------------------------------------------------------------------------
    # Internal function to test whether a function has been called and add it
    #--------------------------------------------------------------------------
    def _check_join_called_first(self, name,endpoint=False):
        if self._qspec.join is not None or len(self._qspec.roots) == 1: return
        if endpoint:
            raise ValueError(("A query over multiple predicates is incomplete without "
                              "'join' clauses connecting these predicates"))
        raise ValueError("A 'join' clause must be specified before '{}'".format(name))

    #--------------------------------------------------------------------------
    # Add a join expression
    #--------------------------------------------------------------------------
    def join(self, *expressions):
        join=process_join(expressions, self._qspec.roots)
        return QueryImpl(self._factmaps, self._qspec.newp(join=join))

    #--------------------------------------------------------------------------
    # Add an order_by expression
    #--------------------------------------------------------------------------
    def where(self, *expressions):
        self._check_join_called_first("where")

        if not expressions:
            self._qspec.newp(where=None)    # Raise an error

        if len(expressions) == 1:
            where = process_where(expressions[0], self._qspec.roots)
        else:
            where = process_where(and_(*expressions), self._qspec.roots)

        nqspec = self._qspec.newp(where=where)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Add an orderered() flag
    #--------------------------------------------------------------------------
    def ordered(self, *expressions):
        self._check_join_called_first("ordered")
        if self._qspec.getp("order_by",None) is not None:
            raise ValueError(("Invalid query 'ordered' declaration conflicts "
                              "with previous 'order_by' declaration"))
        nqspec = self._qspec.newp(ordered=True)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Add an order_by expression
    #--------------------------------------------------------------------------
    def order_by(self, *expressions):
        self._check_join_called_first("order_by")
        if not expressions:
            nqspec = self._qspec.newp(order_by=None)   # raise exception
        elif self._qspec.getp("ordered",False):
            raise ValueError(("Invalid query 'order_by' declaration '{}' "
                              "conflicts with previous 'ordered' "
                              "declaration").format(expressions))
        else:
            nqspec = self._qspec.newp(
                order_by=process_orderby(expressions,self._qspec.roots))
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Add a group_by expression
    #--------------------------------------------------------------------------
    def group_by(self, *expressions):
        if not expressions:
            nqspec = self._qspec.newp(group_by=None)   # raise exception
        else:
            nqspec = self._qspec.newp(
                group_by=process_orderby(expressions,self._qspec.roots))
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Explicitly select the elements to output or delete
    #--------------------------------------------------------------------------
    def select(self,*outsig):
        self._check_join_called_first("select")
        if not outsig:
            raise ValueError("An empty 'select' signature is invalid")
        nqspec = self._qspec.newp(select=outsig)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # The distinct flag
    #--------------------------------------------------------------------------
    def distinct(self):
        self._check_join_called_first("distinct")
        nqspec = self._qspec.newp(distinct=True)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Ground - bind
    #--------------------------------------------------------------------------
    def bind(self,*args,**kwargs):
        self._check_join_called_first("bind")
        nqspec = self._qspec.bindp(*args, **kwargs)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # The tuple flag
    #--------------------------------------------------------------------------
    def tuple(self):
        self._check_join_called_first("tuple")
        nqspec = self._qspec.newp(tuple=True)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # Overide the default heuristic
    #--------------------------------------------------------------------------
    def heuristic(self, join_order):
        nqspec = self._qspec.newp(heuristic=True, joh=join_order)
        return QueryImpl(self._factmaps, nqspec)

    #--------------------------------------------------------------------------
    # End points that do something useful
    #--------------------------------------------------------------------------

    #--------------------------------------------------------------------------
    # For the user to see what the query plan looks like
    #--------------------------------------------------------------------------
    def query_plan(self,*args,**kwargs):
        self._check_join_called_first("query_plan")
        qspec = self._qspec.fill_defaults()

        (factsets,factindexes) = \
            QueryExecutor.get_factmap_data(self._factmaps, qspec)
        return make_query_plan(factindexes.keys(), qspec)

    #--------------------------------------------------------------------------
    # Return the placeholders
    #--------------------------------------------------------------------------
    @property
    def qspec(self):
        return self._qspec

    #--------------------------------------------------------------------------
    # Select to display all the output of the query
    # --------------------------------------------------------------------------
    def all(self):
        self._check_join_called_first("all",endpoint=True)

        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.all()

    #--------------------------------------------------------------------------
    # Show the single element and throw an exception if there is more than one
    # --------------------------------------------------------------------------
    def singleton(self):
        self._check_join_called_first("singleton",endpoint=True)

        qe = QueryExecutor(self._factmaps, self._qspec)
        count = 0
        for out in qe.all():
            count += 1
            if count > 1:
                raise ValueError("Query returned more than a single element")
        if count == 0:
            raise ValueError("Query has no matching elements")
        return out

    #--------------------------------------------------------------------------
    # Return the count of elements - Note: the behaviour of what is counted
    # changes if group_by() has been specified.
    # --------------------------------------------------------------------------
    def count(self):
        self._check_join_called_first("count",endpoint=True)

        qe = QueryExecutor(self._factmaps, self._qspec)

        def group_by_generator():
            for k, g in qe.all():
                yield k, sum(1 for _ in g)

        # If group_by is set then we want to count records associated with each
        # key and not just total records.
        if self._qspec.group_by:
            return group_by_generator()
        else:
            return sum(1 for _ in qe.all())


    #--------------------------------------------------------------------------
    # Return the first element of the query
    # --------------------------------------------------------------------------
    def first(self):
        self._check_join_called_first("first",endpoint=True)

        qe = QueryExecutor(self._factmaps, self._qspec)

        for out in qe.all():
            return out
        raise ValueError("Query has no matching elements")

    #--------------------------------------------------------------------------
    # Delete a selection of fact
    #--------------------------------------------------------------------------
    def delete(self):
        self._check_join_called_first("delete",endpoint=True)

        qe = QueryExecutor(self._factmaps, self._qspec)
        return qe.delete()

#------------------------------------------------------------------------------
# main
#------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError('Cannot run modules')
